{-# OPTIONS_GHC -fglasgow-exts #-} 

-- Code for:
-- 
-- Dagger Traced Monoidal Categories and Reversible Programming.
-- 
-- Copyright (c) 2011, William J. Bowman, Roshan P. James, and Amr
-- Sabry. The code is released under the MIT license.
--
-- This code tested with GHC version 6.12.3

module Pi where

-----------------------------------------------------------------------
-- Isomorphisms 

data Zero 

data a :<=> b where 
-- Congruence
  Id    :: a :<=> a
  Sym   :: (a :<=> b) -> (b :<=> a) 
  (:.:) :: (a :<=> b) -> (b :<=> c) -> (a :<=> c)
  (:*:) :: (a :<=> b) -> (c :<=> d) -> ((a,c) :<=> (b,d))
  (:+:) :: (a :<=> b) -> (c :<=> d) -> (Either a c :<=> Either b d)
-- (+) is associative, commutative, and has a unit
  PlusZeroL   :: Either Zero a :<=> a
  PlusZeroR   :: a :<=> Either Zero a
  CommutePlus :: Either a b :<=> Either b a
  AssocPlusL  :: Either a (Either b c) :<=> Either (Either a b) c 
  AssocPlusR  :: Either (Either a b) c :<=> Either a (Either b c) 
-- (*) is associative, commutative, and has a unit
  TimesOneL    :: ((), a) :<=> a
  TimesOneR    :: a :<=> ((), a)
  CommuteTimes :: (a,b) :<=> (b,a) 
  AssocTimesL  :: (a,(b,c)) :<=> ((a,b),c)
  AssocTimesR  :: ((a,b),c) :<=> (a,(b,c))
-- (*) distributes over (+) 
  TimesZeroL  :: (Zero, a) :<=> Zero
  TimesZeroR  :: Zero :<=> (Zero, a)
  Distribute  :: (Either b c, a) :<=> Either (b, a) (c, a)
  Factor      :: Either (b, a) (c, a) :<=> (Either b c, a)
-- Encoding of booleans
  FoldB   :: Either () () :<=> Bool
  UnfoldB :: Bool :<=> Either () ()
-- Encoding of natural numbers using the isorecursive type (mu x.1+x)
  FoldN   :: Either () Int :<=> Int
  UnfoldN :: Int :<=> Either () Int
-- Encoding of lists of natural numbers using the isorecursive type
-- (mu x.1+nat*x)
  FoldLN :: Either () (Int, [Int]) :<=> [Int]
  UnfoldLN :: [Int] :<=> Either () (Int, [Int])
-- Trance operators for looping/recursion
  TracePlus :: (Either a b1 :<=> Either a b2) -> (b1 :<=> b2)

-- Adjoint

adjoint :: (a :<=> b) -> (b :<=> a)
adjoint Id = Id
adjoint (Sym f) = f 
adjoint (f :.: g) = adjoint g :.: adjoint f
adjoint (f :*: g) = adjoint f :*: adjoint g
adjoint (f :+: g) = adjoint f :+: adjoint g
adjoint PlusZeroL = PlusZeroR
adjoint PlusZeroR = PlusZeroL
adjoint CommutePlus = CommutePlus
adjoint AssocPlusL = AssocPlusR
adjoint AssocPlusR = AssocPlusL
adjoint TimesOneL = TimesOneR
adjoint TimesOneR = TimesOneL
adjoint CommuteTimes = CommuteTimes
adjoint AssocTimesL = AssocTimesR
adjoint AssocTimesR = AssocTimesL
adjoint TimesZeroL = TimesZeroR
adjoint TimesZeroR = TimesZeroL
adjoint Distribute = Factor
adjoint Factor = Distribute
adjoint FoldB = UnfoldB
adjoint UnfoldB = FoldB
adjoint FoldN = UnfoldN
adjoint UnfoldN = FoldN
adjoint FoldLN = UnfoldLN
adjoint UnfoldLN = FoldLN
adjoint (TracePlus c) = TracePlus (adjoint c)

-- Semantics
(@@) :: (a :<=> b) -> a -> b
Id @@ a = a
(Sym f) @@ b = (adjoint f) @@ b
(f :.: g) @@ a = g @@ (f @@ a)
(f :*: g) @@ (a,b) = (f @@ a, g @@ b) 
(f :+: g) @@ (Left a) = Left (f @@ a) 
(f :+: g) @@ (Right b) = Right (g @@ b) 
PlusZeroL @@ (Right a) = a
PlusZeroR @@ a = Right a
CommutePlus @@ (Left a) = Right a
CommutePlus @@ (Right b) = Left b 
AssocPlusL @@ (Left a) = Left (Left a) 
AssocPlusL @@ (Right (Left b)) = Left (Right b) 
AssocPlusL @@ (Right (Right c)) = Right c
AssocPlusR @@ (Left (Left a)) = Left a
AssocPlusR @@ (Left (Right b)) = Right (Left b)
AssocPlusR @@ (Right c) = Right (Right c)
TimesOneL @@ ((), a) = a
TimesOneR @@ a = ((), a)
CommuteTimes @@ (a,b) = (b,a) 
AssocTimesL @@ (a,(b,c)) = ((a,b),c) 
AssocTimesR @@ ((a,b),c)  = (a,(b,c))
Distribute @@ (Left b, a) = Left (b, a) 
Distribute @@ (Right c, a) = Right (c, a) 
Factor @@ (Left (b, a)) = (Left b, a) 
Factor @@ (Right (c, a)) = (Right c, a) 
FoldB @@ (Left ()) = True
FoldB @@ (Right ()) = False
UnfoldB @@ True = Left ()
UnfoldB @@ False = Right ()
FoldN @@ (Left ()) = 0
FoldN @@ (Right n) = 1 + n
UnfoldN @@ 0 = Left ()
UnfoldN @@ n = Right (n-1) 
FoldLN @@ Left () = []
FoldLN @@ Right (h, t) = h : t
UnfoldLN @@ [] = Left ()
UnfoldLN @@ (h : t) = Right (h, t)
(TracePlus c) @@ v = loop c (c @@ (Right v))
    where
      loop c (Left v) = loop c (c @@ (Left v))
      loop c (Right v) = v

------------------------------------------------------------------------
-- Combinational circuits 

-- Not works by first converting the Haskell boolean via a Unfold to
-- either Left () or Right (), where True is Left (), False is Right ().
-- not can simply use CommutePlus to perform a not, then Fold up the
-- bool.
-- REPL Session:
-- *Pi> inot @@ True
-- False
-- *Pi> inot @@ False
inot :: Bool :<=> Bool
inot = UnfoldB :.: CommutePlus :.: FoldB

-- Cond takes two isomorphisms from some type a to some type b, and 
-- creates an isomorphism between a pair of (Bool, a) which will apply
-- the first isomorphism if the Bool is True, and the second if the Bool
-- is False.
cond :: (a :<=> b) -> (a :<=> b) -> ((Bool, a) :<=> (Bool, b))
cond f g = (UnfoldB :*: Id) 
           :.: Distribute 
           :.: ((Id :*: f) :+: (Id :*: g))
           :.: Factor 
           :.: (FoldB :*: Id) 

-- controlled takes an isomorphism between a type a to type a, and 
-- creates an isomorpism (using cond) that will apply the isomorphism to
-- the second value of the pair, if the first value if True, and apply
-- Id otherwise.
controlled :: (a :<=> a) -> ((Bool, a) :<=> (Bool, a))
controlled f = cond f Id

-- cnot is Controlled Not, as found in reversible computing papers such
-- as Reversible Computer by Toffoli. It takes a pair of bools, and
-- applies not to the second Bool if the first is True, and otherwise
-- leaves the value unchanged.
-- REPL Session:
-- *Pi> cnot @@ (True, True)
-- (True, False)
-- *Pi> cnot @@ (False, True)
-- (False, True)
-- *Pi> cnot @@ (True, False)
-- (True, False)
cnot :: (Bool, Bool) :<=> (Bool, Bool)
cnot = controlled inot

-- Toffoli is the universal nand/and gate presented in Reversible
-- Computing by Toffoli.  It is equivalent to controlled controlled
-- not. It takes 3 bools, if the first is True, if applies contolled not
-- to the second 2 bools.
-- REPL Session:
-- *Pi> toffoli @@ ((True, True), False)
-- ((True, True), True)
-- *Pi> toffoli @@ ((True, False), False)
-- ((True, False), False)
-- *Pi> toffoli @@ ((True, True), True)
-- ((True, True), Flase)
-- *Pi> toffoli @@ ((False, True), False)
-- ((False, True), False)
toffoli :: ((Bool,Bool),Bool) :<=> ((Bool,Bool),Bool)
toffoli = AssocTimesR :.: controlled cnot :.: AssocTimesL

-- The Fredkin gate is a well known universal gate.
-- If the first bool is true, it swaps the second two, otherwise it
-- leaves the values unchanged.
fredkin :: (Bool,(Bool,Bool)) :<=> (Bool,(Bool,Bool))
fredkin = controlled CommuteTimes

-- The Peres gate is a universal gate: it takes three inputs a, b, and c, 
-- and produces a, a xor b, (a and b) xor c
peres :: ((Bool,Bool),Bool) :<=> ((Bool,Bool),Bool)
peres = toffoli :.: (cnot :*: Id) 

-- fullAdder can be interpreted as an irreversible 2 bit adder with
-- carry, by fixing the first input to be False and interpretting the
-- inputs and outputs as follows:
--
-- Input: (Constant, ((Number1, Number2), CarryIn)))
-- Output (Garbage1, (Garbage2, (Sum, Carry_Out)))
--
-- All values should be booleans, where False is 0 and True is 1.
-- Constant must be initialized to 0.
fullAdder :: (Bool, ((Bool, Bool), Bool)) :<=> (Bool,(Bool,(Bool,Bool)))
fullAdder = CommuteTimes :.: (CommuteTimes :*: Id) :.: 
            AssocTimesR :.: CommuteTimes :.: (peres :*: Id) :.:
            AssocTimesR :.: (Id :*: CommuteTimes) :.: 
            AssocTimesR :.: (Id :*: AssocTimesL) :.:
            (Id :*: peres) :.: (Id :*: AssocTimesR)

--------------------------------------------------------------
-- Some handy swaps etc.

sw :: (a, (b, c)) :<=> (b, (a, c))
sw = AssocTimesL :.: (CommuteTimes :*: Id) :.: AssocTimesR

sw2 :: ((a, b), (c, d)) :<=> ((a, c), (b, d))
sw2 = AssocTimesR :.: (Id :*: sw) :.: AssocTimesL

-- We can introduce unit freely. hide_unit makes something 
-- that needs a unit temporary value, and creates a new isomorphism that 
-- performs the same function, but automatically introduces and removes
-- a unit temp value.
hide_unit :: (c, ()) :<=> (c, ()) -> c :<=> c
hide_unit c = TimesOneR :.: CommuteTimes :.: c :.: CommuteTimes :.: TimesOneL

------------------------------------------------------------------------
-- Simple primitives on inductive types

-- addSub1 can be thought of as the function add1 mod (the sum of both
-- inputs), by ignoring the second input and output. 
-- 
-- By ignoring the first input and first output, the function can be
-- thought of as sub1 (mod the sum of both inputs)
-- 
-- REPL Session:
-- *Pi> addSub1 @@ (10, 1)
-- (11, 0)
-- *Pi> addSub1 @@ (1, 10)
-- (2, 9)
-- *Pi> addSub1 @@ (10, 0)
-- (0, 10)
addSub1 :: (Int, Int) :<=> (Int, Int)
addSub1 = CommuteTimes 
            :.: (UnfoldN :*: Id) 
            :.: Distribute 
            :.: (Id :+: CommuteTimes)
            :.: Factor 
            :.: (FoldN :*: Id) 

--------------------------------------------------------------------------
-- Iterating a list of nats.

-- iter_ls_nat takes an isomorphism which represents a single step of a
-- loop over a list, and creates an isomorphism which loops over a list,
-- threading through any other arbitrary values.
--
-- To do this, we introduce a unit, which is used to build up a new
-- list, as we traverse the input list.  This is necessary, as information
-- preservation is needed to maintain reversibility.
-- After introducing the unit, we trace a body that does some
-- rearranging and deconstruction of the list.
-- 
-- Step should be an isomorphism that takes a pair, whose first element
-- is a pair of the head and tail of the list at a given step, and whose
-- second element is a pair of the list that is being built to preserve
-- information, and the threaded values of its choosing.
--
-- The resulting isomorphism will take a list and some threaded values,
-- and iterate over the list, performing step each time a tail operation
-- is performed (i.e. the list is 'decremented')
iter_ls_nat :: ((Int, [Int]), ([Int], a)) :<=> ((Int, [Int]), ([Int], a)) -> 
               ([Int], a) :<=> ([Int], a)
iter_ls_nat step = TimesOneR 
                   :.: (TracePlus body)
                   :.: TimesOneL
    where
      -- body :: Either ((Int, [Int]), ([Int], a)) ((), ([Int], a)) :<=> 
      --         Either ((Int, [Int]), ([Int], a)) ((), ([Int], a))
      body = Factor 
             :.: ((CommutePlus :.: FoldLN) :*: Id)
             :.: sw
             :.: ((UnfoldLN :.: CommutePlus) :*: Id)
             :.: Distribute
             :.: ((step :.: sw2) :+: Id)

-- Isomorphisms over lists

-- To reverse a list, we introduce a unit value, as iterating requires
-- (or allows, depending on perspective) that we give some values to
-- thread through the loop, and performs Id at each step.  The resulting
-- list is inherently reversed, due to how the list has to be built up
-- as we iterate it.
--
-- REPL Session:
-- *Pi> ireverse @@ [1..5]
-- [5,4,3,2,1]
-- *Pi> ireverse @@ [5,4..1]
-- [1,2,3,4,5]
ireverse :: [Int] :<=> [Int]
ireverse = hide_unit (iter_ls_nat Id)

-- shuffle performs a shuffle on the list; it reverses the tail of the
-- list at each step of iteration, before recurring on it.
-- *Pi> shuffle @@ [1..5]
-- [1,5,2,4,3]
shuffle :: [Int] :<=> [Int]
shuffle = hide_unit (iter_ls_nat rev') :.: ireverse
    where rev' = (Id :*: ireverse) :*: Id
    
------------------------------------------------------------------------
-- Iterating on a nat.
-- 
-- Given an isomorphsim between a type a, generates a isomorphism 
-- between a pair of an Int and type a, which will apply the given
-- isomorphism at each step, as it iterates over the int.  At each step,
-- the given isomorphsim has access to only the values of a, which are
-- threaded through the loop.
iter_nat ::  (a :<=> a) -> 
             (Int, a) :<=> (Int, a)
iter_nat step = TimesOneR 
                :.: (TracePlus body)
                :.: TimesOneL
    where
      -- body :: Either (Int, (Int, a)) ((), (Int, a))  :<=> 
      --         Either (Int, (Int, a)) ((), (Int, a))
      body = Factor 
             :.: ((CommutePlus :.: FoldN) :*: Id)
             :.: sw
             :.: ((UnfoldN :.: CommutePlus) :*: Id)
             :.: Distribute
             :.: (((Id :*: (Id :*: step)) :.: sw) :+: Id)

-- evenOdd can be thought of as the irreversible function even, by
-- fixing the second input to True, and ignoring the first output. It
-- can also represent the irreversible function odd by fixing the second
-- output to False, and again ignoring the first output.
--
-- REPL Session:
-- *Pi> evenOdd @@ (0, True)
-- (0, True)
-- *Pi> evenOdd @@ (1, True)
-- (1, False)
-- *Pi> evenOdd @@ (0, False)
-- (0, False)
-- *Pi> evenOdd @@ (1, False)
-- (1, True)
-- *Pi> evenOdd @@ (4, False)
-- (4, False)
-- *Pi> evenOdd @@ (5, False)
-- (5, True)
-- *Pi> evenOdd @@ (4, True)
-- (4, True)
evenOdd :: (Int, Bool) :<=> (Int, Bool)
evenOdd = iter_nat inot

-- addSubN can be thought of the irreversible function add by providing
-- h_large for the second input, a first number as the first input, and
-- the second number in the third input, and ignoring the last 2 outputs
--
-- addSubN can be thought of as the irreversible function subtract by
-- providing h_large in the first input, and treating the last two
-- inputs as the inputs to subtract.  By ignoring the first and third
-- outputs, you have the result.
--
-- Note that both addition and subtraction is performed mod the sum of
-- both the first 2 inputs plus 1.  So, if you try to perform 
-- subtraction with h_large = 10, and arguments 0 and 1, the result 
-- will wrap around to 10.
--
-- Sample REPL session:
-- *P> addSubN @@ ((10, 10), 7)
-- ((17,3),7)
-- *P> addSubN @@ ((10, 1000000), 7)
-- ((17,999993),7)
-- *P> addSubN @@ ((10, 1000000), 200)
-- ((210,999800),200)
-- *P> addSubN @@ ((10, 0), 0)
-- ((10,0),0)
-- *P> addSubN @@ ((10, 0), 1)
-- ((0,10),1)
-- *P> addSubN @@ ((10, 0), 2)
-- ((1,9),2)
addSubN :: ((Int, Int), Int) :<=> ((Int, Int), Int)
addSubN = CommuteTimes :.: (iter_nat addSub1) :.: CommuteTimes

-- Mult can be thought of as the irreversible function multiply by
-- fixing the first two arguments to 0 and h_large, respectively, and
-- using the last two inputs as the arguments to multiply.
-- (again, mod the sum of the first two arguments + 1)
--
-- The result is obtained by ignoring the last 3 inputs:
-- 
--
-- mult(((accumulator, heap), n1), n2)
-- Sample REPL Session:
--
-- *P> mult @@ (((0,10000), 2), 3)
-- (((6,9994),2),3)
-- *P> mult @@ (((0,10000), 7), 11)
-- (((77,9923),7),11)
-- *P> mult @@ (((0,10000), 7), 0)
-- (((0,10000),7),0)
-- *P> mult @@ (((0,10000), 0), 11)
-- (((0,10000),0),11)
mult :: (((Int, Int), Int), Int) :<=> (((Int, Int), Int), Int)
mult = CommuteTimes :.: (iter_nat addSubN) :.: CommuteTimes

-- Factorial. Shuffle the accumulator and the 2nd input around.  Assumes the same
-- input as mult.  Used for fact.
fshuf :: (((Int, Int), Int), Int) :<=> (((Int, Int), Int), Int)
fshuf = AssocTimesR 
        :.: AssocTimesR 
        :.: (Id :*: (AssocTimesL :.: CommuteTimes)) 
        :.: AssocTimesL 
        :.: CommuteTimes 
        :.: (CommuteTimes :*: Id) 
        :.: AssocTimesL

-- (((((Acc, Heap), ?), Input), []), [0, 0..]) ->
-- (((((0, Heap), Acc), Input), [?]), [0..])
-- Collect garbage does a massive amount of shuffling, and pushes some
-- garbage onto a garbage list, and pulls a fresh 0 off the 0 list for
-- the new accumulator.
-- It was written in a very systematic way, and as a result is much more
-- verbose than necessary, and rather inefficient.
collect_garbage :: (((((Int, Int), Int), Int), [Int]), [Int]) :<=> 
                   (((((Int, Int), Int), Int), [Int]), [Int])
collect_garbage = (AssocTimesR :*: Id)
                  :.: ((Id :*: CommuteTimes) :*: Id)
                  :.: (AssocTimesR :*: Id)
                  :.: ((Id :*: AssocTimesL) :*: Id)
                  :.: (Id :*: UnfoldLN)
                  :.: CommuteTimes :.: Distribute
                  :.: (CommuteTimes :+: CommuteTimes)
                  :.: (AssocTimesR :+: Id)
                  :.: ((Id :*: (AssocTimesR :*: Id)) :+: (AssocTimesR))
                  :.: ((Id :*: ((Id :*: CommuteTimes) :*: Id)) :+:
                          (Id :*: AssocTimesL))
                  :.: ((Id :*: (AssocTimesL :*: Id)) :+: 
                          (Id :*: ((CommuteTimes :*: Id) :*: Id)))
                  :.: (Id :+: (Id :*: (CommuteTimes :*: Id)))
                  :.: (Id :+: (Id :*: (AssocTimesL :*: Id)))
                  :.: (Id :+: (Id :*: AssocTimesR))
                  :.: (Id :+: (Id :*: (Id :*: CommuteTimes)))
                  :.: (Id :+: (Id :*: AssocTimesL))
                  :.: (AssocTimesL :+: AssocTimesL)
                  :.: (CommuteTimes :+: CommuteTimes)
                  :.: Factor
                  :.: CommuteTimes
                  :.: (AssocTimesR)
                  :.: (Id :*: (Id :*: FoldLN))
                  :.: AssocTimesL
                  :.: (AssocTimesL :*: Id)
                  :.: ((AssocTimesL :*: Id) :*: Id)
                  :.: (AssocTimesR)
                  :.: (Id :*: CommuteTimes)
                  :.: AssocTimesL
                  :.: ((fshuf :*: Id) :*: Id)
-- Reversible fact!
-- Sample Input: ((([], [0,0,0,0,0,0,0]), (((0, 10000), 1), 5)), 5)
-- Sample Output: ((([120,60,20,5,1],[0,0,0,0,0,0]),(((0,9680),120),0)),5)
--
-- Requires a garbage list to store junk in, a (sufficently large) list
-- of zeros, a (sufficiently large) heap number, and 1.  Requires the
-- input twice.  This restriction could be lifted by initilizing one of
-- them to 0 and counting up during the recursion instead of down, but
-- this way it much easier.
--
-- It should also be possible to eliminate one input by using a
-- different sort of iteration over nats, that gives us access to
-- itermediate values.
-- There are many optimizations possible in this code.
fact :: ((([Int], [Int]), (((Int, Int), Int), Int)), Int) :<=>
        ((([Int], [Int]), (((Int, Int), Int), Int)), Int)
fact =  CommuteTimes :.: iter_nat ((Id :*: mult)
                :.: (CommuteTimes :.: AssocTimesL :.: collect_garbage)
                :.: AssocTimesR
                :.: CommuteTimes -- Garbage collected, and accumulator 
                                 -- reinitialized
                :.: (Id :*: (AssocTimesR -- Sub1 from the recursive step
                :.: (Id :*: addSub1)
                :.: AssocTimesL 
                :.: (AssocTimesR :*: Id)
                :.: ((Id :*: addSub1) :*: Id)
                :.: (AssocTimesL :*: Id)
                ))) :.: CommuteTimes

------------------------------------------------------------------------
-- Infinite loops

-- Isomorphic inc function
-- (n, True) --> (n+1, True)
-- (n, False) --> (n-1, False)
-- (0, False) --> (0, True)
-- 
-- the adjoint of this function is also an isomorphism. 
iso_inc :: (Int, Bool) :<=> (Int, Bool)
iso_inc = CommuteTimes 
          :.: (UnfoldB :*: Id)
          :.: Distribute
          :.: (TimesOneL :+: TimesOneL)
          :.: (Id :+: UnfoldN)
          :.: AssocPlusL
          :.: ((CommutePlus :.: FoldN) :+: Id)
          :.: (TimesOneR :+: TimesOneR)
          :.: Factor
          :.: (FoldB :*: Id)
          :.: CommuteTimes

-- inc function
-- n --> n+1
-- 
-- the adjoint of this function is undefined for 0. 
inc :: Int :<=> Int
inc = TracePlus body
    where
      body = ((UnfoldN :.: CommutePlus)  :+: Id)
             :.: AssocPlusR
             :.: (Id :+: FoldN)


-- A partial function that is undefined for all input (n, True) 
-- where n > 0. This function is a partial surjection. Its inverse 
-- is a total function.
omega0 :: (Int, Bool) :<=> (Int, Bool)
omega0 = TimesOneR
         :.: (TracePlus body)
         :.: TimesOneL
    where
      -- body :: Either (Int, (Int, a)) ((), (Int, a))  :<=>
      --         Either (Int, (Int, a)) ((), (Int, a))
      body = Factor
             :.: ((CommutePlus :.: FoldN) :*: Id)
             :.: sw
             :.: ((UnfoldN :.: CommutePlus) :*: Id)
             :.: Distribute
             :.: ((sw :.: (Id :*: iso_inc)) :+: Id)


-- A couple of functions based on how we compose the above. 
--
-- This is a partial function. It is the identity on the defined
-- inputs.  i.e. (c v) |-->* v if it terminates.
omega0_partial_id = omega0 :.: (adjoint omega0)

-- This is just the identity. 
omega0_id = (adjoint omega0) :.: omega0

-- Another infinite loop, but this time on a finite type.  This is
-- defined only on input False. Input True causes it to diverge. Since
-- this is on a finite type, we can ask our usual information
-- theoretic questions about this: Does this function preserve
-- information?
omega1 :: Bool :<=> Bool
omega1 = TimesOneR 
         :.: (TracePlus body)
         :.: TimesOneL
    where
      -- body :: Either (Int, a) ((), a)  :<=> 
      --         Either (Int, a) ((), a)
      body = Factor 
             :.: ((CommutePlus :.: FoldN) :*: Id)
             :.: iso_inc
             :.: ((UnfoldN :.: CommutePlus) :*: Id)
             :.: Distribute


-- Undefined on all inputs. Does this constitute some form of
-- "deleting a bit"? Entropy of bool is 1. 
omega1_bool :: Bool :<=> Bool
omega1_bool = omega1 :.: omega1

-- Another infinite loop.  Unit however is supposed to have no
-- information, so what does non-termination mean in this case?
-- Entropy of unit is 0.
omega1_unit :: () :<=> ()
omega1_unit = TracePlus (FoldB :.: omega1 :.: UnfoldB)

------------------------------------------------------------------------
